#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
import os
import shutil
try:
    import magic
except ImportError:
    magic = None

from io import BytesIO
from PIL import Image

# import interfaces
from pyams_file.interfaces import IFile, IImage, IVideo, IAudio, IFileInfo, FileModifiedEvent
from zope.copy.interfaces import ICopyHook, ResumeCopy
from zope.location.interfaces import IContained

# import packages
from persistent import Persistent
from pyams_utils.adapter import adapter_config, ContextAdapter
from pyams_utils.request import check_request
from ZODB.blob import Blob
from zope.container.contained import Contained
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty


BLOCK_SIZE = 1 << 16


@implementer(IFile, IFileInfo, IContained)
class File(Persistent, Contained):
    """Generic file persistent object"""

    title = FieldProperty(IFileInfo['title'])
    description = FieldProperty(IFileInfo['description'])
    filename = FieldProperty(IFileInfo['filename'])
    language = FieldProperty(IFileInfo['language'])

    def __init__(self, data='', content_type=None, source=None):
        self.content_type = content_type
        self._blob = None
        if data:
            self.data = data
        elif source:
            if os.path.exists(source):
                try:
                    f = open(source, 'rb')
                    self.data = f
                finally:
                    f.close()

    def get_blob(self, mode='r'):
        if self._blob is None:
            return None
        return self._blob.open(mode=mode)

    def get_detached_blob(self):
        if self._blob is None:
            return None
        return open(self._blob.committed(), 'rb')

    def _get_data(self):
        f = self.get_blob()
        if f is None:
            return None
        try:
            data = f.read()
            return data
        finally:
            f.close()

    def _set_data(self, data):
        if self._blob is None:
            self._blob = Blob()
        if isinstance(data, str):
            data = data.encode('utf-8')
        elif hasattr(data, 'seek'):
            data.seek(0)
        f = self.get_blob('w')
        try:
            if hasattr(data, 'read'):
                self._size = 0
                _data = data.read(BLOCK_SIZE)
                size = len(_data)
                while size > 0:
                    f.write(_data)
                    self._size += size
                    _data = data.read(BLOCK_SIZE)
                    size = len(_data)
            else:
                f.write(data)
                self._size = len(data)
        finally:
            f.close()

    data = property(_get_data, _set_data)

    def get_size(self):
        return self._size

    def __enter__(self):
        return self.get_blob(mode='c')

    def __exit__(self, exc_type, exc_val, exc_tb):
        # exc_val.value.close()
        pass

    def __iter__(self):
        if self._blob is None:
            raise StopIteration()
        with self as f:
            while True:
                chunk = f.read(BLOCK_SIZE)
                if not chunk:
                    raise StopIteration(f)
                yield chunk

    def __nonzero__(self):
        return self._size > 0


@implementer(IImage)
class ImageFile(File):
    """Image file persistent object"""

    image_size = (-1, -1)

    def _set_data(self, data):
        if isinstance(data, str):
            data = BytesIO(data.encode('utf-8'))
        elif isinstance(data, bytes):
            data = BytesIO(data)
        File._set_data(self, data)
        if hasattr(data, 'seek'):
            data.seek(0)
        img = Image.open(data)
        self.image_size = img.size

    data = property(File._get_data, _set_data)

    def get_image_size(self):
        return self.image_size

    def resize(self, width, height, keep_ratio=True):
        image = Image.open(self.get_blob(mode='c'))
        image_size = image.size
        if width >= image_size[0] and height >= image_size[1]:
            return
        new_image = BytesIO()
        w_ratio = 1. * width / image_size[0]
        h_ratio = 1. * height / image_size[1]
        if keep_ratio:
            ratio = min(w_ratio, h_ratio)
            image.resize((round(ratio * image_size[0]), round(ratio * image_size[1])), Image.ANTIALIAS) \
                 .save(new_image, image.format, quality=99)
        else:
            image.resize((round(w_ratio * image_size[0]), round(h_ratio * image_size[1])), Image.ANTIALIAS) \
                 .save(new_image, image.format, quality=99)
        self.data = new_image
        request = check_request()
        request.registry.notify(FileModifiedEvent(self))

    def crop(self, x1, y1, x2, y2):
        image = Image.open(self.get_blob(mode='c'))
        new_image = BytesIO()
        image.crop((x1, y1, x2, y2)) \
             .save(new_image, image.format, quelity=99)
        self.data = new_image
        request = check_request()
        request.registry.notify(FileModifiedEvent(self))


@implementer(IVideo)
class VideoFile(File):
    """Video file persistent object"""


@implementer(IAudio)
class AudioFile(File):
    """Audio file persistent object"""


def get_magic_content_type(input):
    """Get content-type based on magic library"""
    if magic is not None:
        if hasattr(input, 'seek'):
            input.seek(0)
        if hasattr(input, 'read'):
            input = input.read()
        return magic.from_buffer(input, mime=True)
    else:
        return None


def FileFactory(data):
    """File object factory

    Automatically create the right file type based on magic
    content-type recognition
    """
    content_type = get_magic_content_type(data)
    if content_type.startswith(b'image/'):
        factory = ImageFile
    elif content_type.startswith(b'video/'):
        factory = VideoFile
    elif content_type.startswith(b'audio/'):
        factory = AudioFile
    else:
        factory = File
    return factory(data, content_type)


@adapter_config(context=IFile, provides=ICopyHook)
class BlobFileCopyHook(ContextAdapter):
    """Blob file copy hook

    Copied from z3c.blobfile package
    """

    def __call__(self, toplevel, register):
        register(self._copy_blob)
        raise ResumeCopy

    def _copy_blob(self, translate):
        blob = self.context._blob
        if blob is not None:
            target = translate(self.context)
            target._blob = Blob()
            fsrc = blob.open('r')
            fdst = target._blob.open('w')
            shutil.copyfileobj(fsrc, fdst)
            fdst.close()
            fsrc.close()
